// Copyright 2022 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

package configuration

import (
	"bytes"
	"encoding/json"
	"reflect"

	"github.com/imdario/mergo"
	"github.com/magiconair/properties"
	"github.com/pelletier/go-toml/v2"
	"gopkg.in/yaml.v3"
	"io"
	"os"
	"path"
)

func importJson[C any](config *C, reader io.Reader) error {
	decoder := json.NewDecoder(reader)
	return decoder.Decode(config)
}

func importYaml[C any](config *C, reader io.Reader) error {
	decoder := yaml.NewDecoder(reader)
	return decoder.Decode(config)
}

func importToml[C any](config *C, reader io.Reader) error {
	decoder := toml.NewDecoder(reader)
	return decoder.Decode(config)
}

func importProperties[C any](config *C, reader io.Reader) error {

	buf := &bytes.Buffer{}
	buf.ReadFrom(reader)
	b := buf.Bytes()

	p, err := properties.Load(b, properties.UTF8)
	if err != nil {
		return err
	}

	return p.Decode(config)

}

func (s *Settings[C]) importStream(r reader, f ...func(n *C)) {
	var c C
	var err error

	defer func() {
		s.notifyErrorHooks()
	}()

	err = nil // reset error

	x := r.reader
	switch r.format {
	case Json:
		err = importJson(&c, x)
	case Yaml:
		err = importYaml(&c, x)
	case Toml:
		err = importToml(&c, x)
	case Properties:
		err = importProperties(&c, x)
	default:
		err = newFormatNotSupportedError(r.format)
	}

	if err == nil {
		s.importCounter++
	} else {
		s.errors = append(s.errors, err)
	}

	if f != nil {
		for _, fn := range f {
			fn(&c)
		}
	}

	if err := mergo.Merge(&s.config, c); err != nil {
		s.errors = append(s.errors, err)
	}
}

func (s *Settings[C]) importStreams() {
	for _, r := range s.stream.readers {
		s.importStream(r)
	}
}

func replacePath(p string, c interface{}) {
	cValue := reflect.ValueOf(c)

	// If c is of type PathValue (which implements pathInterface), modify the path
	if cValue.Type() == reflect.TypeOf(PathValue("")) {
		pathVal := cValue.Interface().(PathValue)
		if !path.IsAbs(pathVal.String()) {
			newPath := PathValue(path.Join(p, pathVal.String()))
			if cValue.CanSet() {
				cValue.Set(reflect.ValueOf(newPath))
			}
		}
		return
	}

	// If c is not a struct, simply return
	if cValue.Kind() != reflect.Ptr || cValue.Elem().Kind() != reflect.Struct {
		return
	}

	fields := reflect.VisibleFields(cValue.Elem().Type())
	for _, field := range fields {
		r := cValue.Elem().FieldByName(field.Name)
		handleField(p, r)
	}
}

func handleField(p string, r reflect.Value) {
	switch r.Kind() {
	case reflect.Struct:
		if r.CanAddr() {
			replacePath(p, r.Addr().Interface())
		}
	case reflect.Slice:
		for i := 0; i < r.Len(); i++ {
			elem := r.Index(i)
			if elem.Type() == reflect.TypeOf(PathValue("")) {
				pathVal := elem.Interface().(PathValue)
				if !path.IsAbs(pathVal.String()) {
					newPath := PathValue(path.Join(p, pathVal.String()))
					if elem.CanSet() {
						elem.Set(reflect.ValueOf(newPath))
					}
				}
			} else if elem.CanAddr() {
				replacePath(p, elem.Addr().Interface())
			}
		}
	case reflect.Map:
		for _, k := range r.MapKeys() {
			elem := r.MapIndex(k)
			if elem.Type() == reflect.TypeOf(PathValue("")) {
				pathVal := elem.Interface().(PathValue)
				if !path.IsAbs(pathVal.String()) {
					newPath := PathValue(path.Join(p, pathVal.String()))
					r.SetMapIndex(k, reflect.ValueOf(newPath))
				}
			} else if elem.CanAddr() {
				replacePath(p, elem.Addr().Interface())
			}
		}
	default:
		// Check for pathInterface
		if v, ok := r.Interface().(pathInterface); ok {
			// Check if r is nil
			if r.Kind() == reflect.Ptr && r.IsNil() {
				return
			}
			currentPath := v.String()
			if r.CanSet() && !path.IsAbs(currentPath) {
				if r.Type() == reflect.TypeOf(PathValue("")) {
					r.Set(reflect.ValueOf(PathValue(path.Join(p, currentPath))))
				}
			}
		}
	}
}

func (s *Settings[C]) importFiles() {

        defer func() {
		s.notifyErrorHooks()
	}()

	for _, d := range s.files.directories {
		fn := path.Join(d, s.files.name+s.files.format.Extension())

		f, err := s.files.fs.Open(fn)

		if os.IsNotExist(err) {
			_ = f.Close()
			continue
		}

		r := (io.Reader)(f)
		s.importStream(reader{s.files.format, r}, func(c *C) {
			replacePath(d, c)
		})
		_ = f.Close()
	}

	for _, f := range s.files.files {
		r, err := s.files.fs.Open(f.path)

		if err != nil {
			s.errors = append(s.errors, err)
			_ = r.Close()
			continue
		}

		s.importStream(reader{f.format, r}, func(c *C) {
			d := path.Dir(f.path)
			replacePath(d, c)
		})

		_ = r.Close()
	}

}

func (s *Settings[C]) Import() *Settings[C] {

	s.Lock()

	defaults := s.config

	var n C
	s.config = n

	s.importCounter = 0

	s.importFiles()
	s.importStreams()

	defer func() {
		s.notifyErrorHooks()
	}()

	if err := mergo.Merge(&s.config, defaults); err != nil {
		s.errors = append(s.errors, err)
	}

	if s.importCounter == 0 {
		s.errors = append(s.errors, NoFileImportedError)
	}

	x := s.config
	s.config = defaults
	_, l := s.setConfigInternal(x)
	s.Unlock()
	s.notifyHooks(l)

	return s
}
